SystemOrganization addCategory: #'System-History'!

Object subclass: #HistoryIterator
	instanceVariableNames: 'index plugged recorder maxSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-History'!

!HistoryIterator commentStamp: 'AlainPlantec 12/14/2010 23:40' prior: 0!
A HistoryIterator holds an UndoRedoGroup in order to store an history of commands with the possibility of undoing and redoing. The iterator cursor is represented by the index inst var. Index always contains the position of the command that can be currently undone. So, undo decrease index and redo increase index. When a new record is stored, then, index contains the newly added record position.
See HistoryIteratorTest for examples.

Instance Variables
	index:		<Integer>
	maxSize:		<Integer>
	plugged:		<Boolean>
	recorder:		<UndoRedoGroup>

index
	- the iterator cursor

maxSize
	- the maximum number of records that can be added in the root group.

plugged
	- if false, then adding of command is not allowed. Useful to prevent bad history recording recurssions (record while undoing or redoing).

recorder
	- The root of the history tree which records undo/redo commands
!

!HistoryIterator methodsFor: 'adding' stamp: 'AlainPlantec 12/14/2010 18:52'!
addItem: anHistoryItem
	| result |
	self isPlugged ifFalse: [^ false].
	self recorder removeLast: (self size - self index).
	result := self recorder addItem: anHistoryItem.
	self updateIndex.
	^ result! !

!HistoryIterator methodsFor: 'adding' stamp: 'AlainPlantec 12/14/2010 12:54'!
addRecord: anItem
	^ self addItem: anItem
! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 19:03'!
at: anInteger
	^ self recorder at: anInteger! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 19:59'!
at: anInteger put: anItem
	self recorder at: anInteger put: anItem! !

!HistoryIterator methodsFor: 'grouping' stamp: 'AlainPlantec 12/14/2010 14:33'!
closeGroup
	self recorder closeGroup
! !

!HistoryIterator methodsFor: 'compatibility' stamp: 'AlainPlantec 12/14/2010 12:54'!
closeRecordGroup
	^ self closeGroup! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 13:00'!
current
	^ (self index <= self size and: [self index > 0])
		ifTrue: [self recorder at: self index]! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 18:49'!
defaultMaximumSize
	"unlimited by default"
	^ 999999999! !

!HistoryIterator methodsFor: 'undo-undo' stamp: 'AlainPlantec 12/14/2010 18:31'!
do
	^ self redo! !

!HistoryIterator methodsFor: 'undo-undo' stamp: 'AlainPlantec 12/14/2010 22:27'!
doAndAddRecord: anUndoRedoRecord
	| result |
	result := anUndoRedoRecord do.
	self addItem: anUndoRedoRecord.
	^ result! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 19:08'!
first
	^ self recorder first! !

!HistoryIterator methodsFor: 'grouping' stamp: 'MarcusDenker 5/31/2011 15:33'!
groupFrom: firstIdx to: secondIdx
	| group |
	group := UndoRedoGroup new.
	firstIdx to: (secondIdx min: index) do: [:i | group addItem: (self at:i)].
	group close.
	group 
		isEmpty ifTrue: [^ self].
	firstIdx + 1 to: (secondIdx min: index) do: [:i | self removeAt: i].
	self at: firstIdx put: group.
	index := index - group size + 1.! !

!HistoryIterator methodsFor: 'testing' stamp: 'AlainPlantec 12/14/2010 12:58'!
hasNext
	"is there an item after current index"
	^ self recorder size - self index > 0! !

!HistoryIterator methodsFor: 'testing' stamp: 'AlainPlantec 12/14/2010 12:54'!
hasPrevious
	"is there an item before current index"
	^ self index > 0! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 12:54'!
index
	^ index ifNil: [index := self size]! !

!HistoryIterator methodsFor: 'testing' stamp: 'AlainPlantec 12/14/2010 12:54'!
isPlugged
	^ plugged ifNil: [plugged := true]! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 19:06'!
last
	^ self recorder last! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 18:46'!
maxSize
	^ maxSize ifNil: [maxSize := self defaultMaximumSize]! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 18:54'!
maxSize: anInteger
	maxSize := anInteger! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 12:54'!
next
	"next item in history"
	^ self hasNext
		ifTrue: [ index := self index + 1.
			self current]! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 13:01'!
nextList
	"return 'next' items sublist "
	^ self recorder copyFrom: self index + 1 to: self size! !

!HistoryIterator methodsFor: 'grouping' stamp: 'AlainPlantec 12/14/2010 18:52'!
openGroup
	self recorder openGroup.
	self updateIndex.
! !

!HistoryIterator methodsFor: 'compatibility' stamp: 'AlainPlantec 12/14/2010 12:54'!
openRecordGroup
	^ self openGroup! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 12:54'!
previous
	"previous item in history"
	^ self hasPrevious
		ifTrue: [index := self index - 1.
			self current]! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 13:01'!
previousList
	"return 'prev' items sublist"
	^ self recorder copyFrom: 1 to: self index - 1! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 14:37'!
recorder
	^ recorder ifNil: [recorder := UndoRedoGroup new]! !

!HistoryIterator methodsFor: 'undo-undo' stamp: 'AlainPlantec 12/14/2010 13:03'!
redo
	self unplugWhile: [
		self hasNext ifFalse: [^false].
		self next redo].
	^ true! !

!HistoryIterator methodsFor: 'undo-undo' stamp: 'AlainPlantec 12/14/2010 22:27'!
redo: doMessageSend undo: undoMessageSend
	self addItem: (UndoRedoRecord redo: doMessageSend undo: undoMessageSend)! !

!HistoryIterator methodsFor: 'undo-undo' stamp: 'AlainPlantec 12/14/2010 22:28'!
redoArray: doArray undoArray: undoArray
	self addItem: (UndoRedoRecord redoArray: doArray undoArray: undoArray)! !

!HistoryIterator methodsFor: 'removing' stamp: 'AlainPlantec 12/14/2010 19:58'!
removeAt: anIndex
	self recorder removeAt: anIndex! !

!HistoryIterator methodsFor: 'removing' stamp: 'AlainPlantec 12/14/2010 18:51'!
removeFirst
	self recorder removeFirst! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 13:02'!
reset
	self recorder reset.
	index := nil.! !

!HistoryIterator methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 13:00'!
size
	^ self recorder size! !

!HistoryIterator methodsFor: 'undo-undo' stamp: 'AlainPlantec 12/14/2010 13:03'!
undo
	self unplugWhile: [
		self current ifNotNil: [:curr | curr undo].
		self previous].
	^ true! !

!HistoryIterator methodsFor: 'undo-undo' stamp: 'AlainPlantec 12/14/2010 12:54'!
unplugWhile: aBlock
	| wasPlugged |
	wasPlugged := self isPlugged.
	plugged := false.
	aBlock ensure: [ plugged := wasPlugged ]! !

!HistoryIterator methodsFor: 'adding' stamp: 'AlainPlantec 12/14/2010 18:51'!
updateIndex
	self size <= self maxSize
		ifFalse: [self removeFirst].
	index := self size.
! !

Object subclass: #HistoryLeaf
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-History'!

!HistoryLeaf commentStamp: 'AlainPlantec 12/14/2010 23:06' prior: 0!
This class represents the abstract leaf structure of items stored in History tree. Composite nodes of such trees are made of HistoryNode instances. For a particular use, this class need to be subclassed. See for example UndoRedoRecord!

!HistoryLeaf methodsFor: 'adding' stamp: 'AlainPlantec 12/13/2010 13:07'!
addItem: anHistoryItem 
	^ false! !

!HistoryLeaf methodsFor: 'testing' stamp: 'AlainPlantec 12/13/2010 13:04'!
isComposite
	^ false! !

!HistoryLeaf methodsFor: 'testing' stamp: 'AlainPlantec 12/13/2010 11:54'!
opened
	^ false! !

HistoryLeaf subclass: #HistoryNode
	instanceVariableNames: 'history opened'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-History'!

!HistoryNode commentStamp: 'AlainPlantec 12/15/2010 00:04' prior: 0!
A HistoryNode is composite node of an history tree. It is made to contain some other HistoryNode or HistoryLeaf instances.
A subnode is added with #addItem:
In order to add and feed a new subtree, one can use openGroup which add a new HistoryNode. When openGroup is sent to an HistoryNode named H, then a new group G is added and all subsequent sent of #addItem: or of #openGroup to H will update the new node G until G is closed by a closeGroup. 

As examples:
---------------
H := HistoryNode new.
H addItem: (i1 := HistoryLeaf new).
---------------
gives:
H
	i1

---------------
H := HistoryNode new.
H openGroup. "add a new group named g1"
H addItem: (i1 := HistoryLeaf new).
H addItem: (i2 := HistoryLeaf new).
--------------
gives:
H
	g1
		i1
		i2

--------------
H := HistoryNode new.
H openGroup. "add a new group named g1"
H openGroup. "add a new group named g2"
H addItem: (i1 := HistoryLeaf new).
H addItem: (i2 := HistoryLeaf new).
--------------
gives:
H
	g1
		g2
			i1
			i2
			
--------------
H := HistoryNode new.
H openGroup. "add a new group named g1"
H openGroup. "add a new group named g2"
H addItem: (i1 := HistoryLeaf new).
H closeGroup. "close g2"
H addItem: (i2 := HistoryLeaf new).
H closeGroup. "close g1"
H addItem: (i3 := HistoryLeaf new).
--------------
gives:
H
	g1
		g2
			i1
		i2
	i3
	
Also se HistoryNodeTest.


Instance Variables
	history:		<OrderedCollection>
	opened:		<Boolean>

history
	- The list of subnodes (HistoryLeaf or HistoryNode instances)

opened
	- true if the node is opened for adding
!

!HistoryNode methodsFor: 'adding' stamp: 'AlainPlantec 12/14/2010 18:49'!
addItem: anHistoryItem 
	self current notNil
		ifTrue: [(self current addItem: anHistoryItem)
				ifTrue: [^ true]].
	self opened
		ifTrue: [self history add: anHistoryItem.
			^ true].
	^ false! !

!HistoryNode methodsFor: 'accessing' stamp: 'AlainPlantec 12/12/2010 19:01'!
at: aPosition
	^ self history at: aPosition! !

!HistoryNode methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 19:42'!
at: aPosition put: anItem
	self history at: aPosition put: anItem! !

!HistoryNode methodsFor: 'opening-closing' stamp: 'AlainPlantec 12/13/2010 11:53'!
close
	opened := false! !

!HistoryNode methodsFor: 'opening-closing' stamp: 'AlainPlantec 1/5/2011 09:36'!
closeGroup
	(self current notNil
			and: [self current isComposite
				and: [self current opened]])
		ifTrue: [self current closeGroup]
		ifFalse: [self close]! !

!HistoryNode methodsFor: 'testing' stamp: 'AlainPlantec 12/14/2010 19:04'!
closed
	^ self opened not! !

!HistoryNode methodsFor: 'copying' stamp: 'AlainPlantec 12/14/2010 13:06'!
copyFrom: start to: stop
	^ self history copyFrom: start to: stop! !

!HistoryNode methodsFor: 'accessing' stamp: 'AlainPlantec 12/13/2010 15:05'!
current
	^ self history isEmpty
		ifFalse: [self history last] ! !

!HistoryNode methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 19:08'!
first
	^ self history first! !

!HistoryNode methodsFor: 'accessing' stamp: 'AlainPlantec 12/13/2010 10:39'!
groupClass
	^ self class! !

!HistoryNode methodsFor: 'accessing' stamp: 'AlainPlantec 12/12/2010 18:37'!
history
	^ history ifNil: [history := OrderedCollection new] ! !

!HistoryNode methodsFor: 'testing' stamp: 'AlainPlantec 12/13/2010 13:04'!
isComposite
	^ true! !

!HistoryNode methodsFor: 'testing' stamp: 'AlainPlantec 12/13/2010 15:59'!
isEmpty
	^ self history isEmpty! !

!HistoryNode methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 19:08'!
last
	^ self history last! !

!HistoryNode methodsFor: 'opening-closing' stamp: 'AlainPlantec 12/13/2010 11:53'!
open
	opened := true! !

!HistoryNode methodsFor: 'opening-closing' stamp: 'AlainPlantec 12/13/2010 13:07'!
openGroup
	^ self addItem: self groupClass new! !

!HistoryNode methodsFor: 'testing' stamp: 'AlainPlantec 12/13/2010 11:54'!
opened
	^ opened ifNil: [opened := true]! !

!HistoryNode methodsFor: 'removing' stamp: 'AlainPlantec 12/14/2010 19:57'!
removeAt: anIndex
	self history removeAt: anIndex! !

!HistoryNode methodsFor: 'private ' stamp: 'AlainPlantec 12/12/2010 19:06'!
removeFirst
	self history removeFirst! !

!HistoryNode methodsFor: 'removing' stamp: 'AlainPlantec 12/14/2010 12:56'!
removeLast
	self history removeLast! !

!HistoryNode methodsFor: 'removing' stamp: 'AlainPlantec 12/14/2010 12:59'!
removeLast: count
	self history removeLast: count! !

!HistoryNode methodsFor: 'opening-closing' stamp: 'AlainPlantec 12/13/2010 22:00'!
reset
	history := nil.
	opened := true! !

!HistoryNode methodsFor: 'accessing' stamp: 'AlainPlantec 12/12/2010 19:03'!
size
	^ self history size! !

HistoryNode subclass: #UndoRedoGroup
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-History'!

!UndoRedoGroup commentStamp: 'AlainPlantec 12/14/2010 23:30' prior: 0!
An UndoRedoGroup is a composite node of an history tree. It is made to allow redoing or undoing of several commands in one shot. See for example its use to record all text replacement when a "replace all" is performed from a FindReplaceService (see TextEditor>>replaceAll: with: startingAt:)

Instance Variables
!

!UndoRedoGroup methodsFor: 'undo-undo' stamp: 'AlainPlantec 12/14/2010 18:31'!
do
	self redo! !

!UndoRedoGroup methodsFor: 'undo-undo' stamp: 'AlainPlantec 12/14/2010 13:18'!
redo
	self history do: [:i | i redo]! !

!UndoRedoGroup methodsFor: 'undo-undo' stamp: 'AlainPlantec 12/14/2010 19:29'!
undo
	self history reverseDo: [:i | i undo]! !

HistoryLeaf subclass: #UndoRedoRecord
	instanceVariableNames: 'redoMessage undoMessage'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-History'!

!UndoRedoRecord commentStamp: 'AlainPlantec 12/14/2010 23:26' prior: 0!
An UndoRedoRecord is a leaf of an history tree. It  holds undo and redo commands. Such a command is represented by a MessageSend. 

Instance Variables
	redoMessage:		<MessageSend>
	undoMessage:		<MessageSend>

redoMessage
	- The command which is performed for redoing

undoMessage
	- The command which is performed for undoing
!

!UndoRedoRecord class methodsFor: 'compatibility' stamp: 'AlainPlantec 12/14/2010 08:15'!
do: doMessageSend undo: undoMessageSend
	^ self redo: doMessageSend undo: undoMessageSend! !

!UndoRedoRecord class methodsFor: 'compatibility' stamp: 'AlainPlantec 12/14/2010 08:15'!
doArray: doArray undoArray: undoArray
	^ self redoArray: doArray undoArray: undoArray! !

!UndoRedoRecord class methodsFor: 'instance creation' stamp: 'AlainPlantec 12/13/2010 22:43'!
redo: doMessageSend undo: undoMessageSend

	| anUndoRecord |
	anUndoRecord := self new.
	anUndoRecord redoMessage: doMessageSend.
	anUndoRecord undoMessage: undoMessageSend.
	^anUndoRecord.! !

!UndoRedoRecord class methodsFor: 'instance creation' stamp: 'AlainPlantec 12/13/2010 22:43'!
redoArray: doArray undoArray: undoArray
	" doArray and undoArray are constructed as follow: #(receiver selector arguments) "
	| anUndoRecord |
	anUndoRecord := self new.
	anUndoRecord redoMessage: (MessageSend 
		receiver: 	(doArray first) 
		selector: 	(doArray second) 
		arguments: 	(doArray third)).
	undoArray 
		ifNotNil: [ 
			anUndoRecord undoMessage: (MessageSend 
				receiver: 	(undoArray first) 
				selector: 	(undoArray second) 
				arguments: 	(undoArray third))].
	^anUndoRecord.! !

!UndoRedoRecord methodsFor: 'compatibility' stamp: 'AlainPlantec 12/14/2010 08:16'!
do
	^self redo! !

!UndoRedoRecord methodsFor: 'compatibility' stamp: 'AlainPlantec 12/14/2010 08:16'!
doMessage

	^ self redoMessage! !

!UndoRedoRecord methodsFor: 'compatibility' stamp: 'AlainPlantec 12/14/2010 08:17'!
doMessage: aMessageSend
	^ self redoMessage: aMessageSend
! !

!UndoRedoRecord methodsFor: 'redo-undo' stamp: 'AlainPlantec 12/14/2010 13:12'!
redo
	^self redoMessage value.! !

!UndoRedoRecord methodsFor: 'accessing' stamp: 'AlainPlantec 12/13/2010 22:42'!
redoMessage

	^ redoMessage! !

!UndoRedoRecord methodsFor: 'accessing' stamp: 'AlainPlantec 12/13/2010 22:42'!
redoMessage: aMessageSend

	redoMessage := aMessageSend! !

!UndoRedoRecord methodsFor: 'redo-undo' stamp: 'AlainPlantec 12/14/2010 13:12'!
undo
	^self undoMessage
		ifNotNil: [self undoMessage value].! !

!UndoRedoRecord methodsFor: 'accessing' stamp: 'AlainPlantec 12/13/2010 22:22'!
undoMessage

	^ undoMessage! !

!UndoRedoRecord methodsFor: 'accessing' stamp: 'AlainPlantec 12/13/2010 22:22'!
undoMessage: aMessageSend

	undoMessage := aMessageSend! !
